#!/usr/bin/env python
"""
6.100A Spring 2023: Recitation 3
Author: Matthew Feng (mattfeng@mit.edu)
"""

import copy
import timeit
from collections import defaultdict

def list_comprehensions():
    def make_abbreviation(string):
        return "".join(w[0] for w in string.split())

    def make_abbreviation_case(string):
        return "".join(w[0] for w in string.split() if w[0].isupper())

    print(make_abbreviation("United States of America"))
    print(make_abbreviation_case("United States of America"))

def functions_as_arguments():
    def add_two(x):
        return x + 2

    list1 = [1, 2, 3, 4]
    list1_plus2 = list(map(add_two, list1))
    print(list1, list1_plus2)

    # get the length of every element in a list
    L = "You've survived week three!".split() # splits on all whitespace
    print(L)
    print(list(map(len, L)))

def anonymous_functions():
    add_two = lambda x: x + 2
    print(add_two(100))

    f = lambda a, b: (a + b, a - b)
    print(f(5, 8))

    T = [("Matt", 3), ("Jenny", 1), ("Ryan", 2)]
    T_sorted = sorted(T, key=lambda x: x[1])
    print(T_sorted)

def dict_basics():
    name_to_department = {
        "Amy": "DMSE",
        "Rachel": "Architecture",
        "Matt": "EECS",
        "John": "Mathematics"
    }

    name_to_department["Chris"] = "Biology"
    del name_to_department["John"]

    print(name_to_department["Amy"])

    # iterating over a dictionary
    for name, department in name_to_department.items():
        print(f"{name} is in the {department} department.")
    
    # defaultdict is convenient if you want to initialize new entries
    # to a particular value
    registration = defaultdict(list)
    # registration["Matt"].append("Symmetry for ML")
    # registration["Matt"].append("Linear Algebra")
    registration["Matt"].extend(["Symmetry for ML", "Linear Algebra"])
    # registration["Matt"].extend("Symmetry for ML", "Linear Algebra") # wrong!
    print(registration["Matt"])

def set_examples():
    squares_list = list(x ** 2 for x in range(100000))
    squares_set = set(x ** 2 for x in range(100000))

    list_time = timeit.timeit(lambda: 71690086 in squares_list, number=10000)
    set_time = timeit.timeit(lambda: 71690086 in squares_set, number=10000)

    print(f"list: {list_time} ms")
    print(f"set: {set_time} ms")

def mutability_demo():
    def print_ttt_board(rep):
        for row in rep:
            print("".join(row))
        print()

    def list_representation():
        print("### list representation ###")
        undos = []
        board = [
            list("..x"),
            list(".o."),
            list("...")
        ]
        print_ttt_board(board)

        def update_board(board, row, col, value, undos):
            # undos.append(board) # incorrect! needs to use deepcopy
            undos.append(copy.deepcopy(board))

            board[row][col] = value

        update_board(board, 2, 0, "x", undos)
        print_ttt_board(board)
        print_ttt_board(undos[-1])
    
    def tuple_representation():
        print("### tuple representation ###")
        undos = []
        board = (
            tuple("..x"),
            tuple(".o."),
            tuple("...")
        )
        print_ttt_board(board)

        def update_board(board, row, col, value, undos):
            undos.append(board)
            new_board = (
                (
                    value if r == row and c == col else board[r][c]
                    for c in range(3)
                )
                for r in range(3)
            )

            return new_board

        board = update_board(board, 2, 0, "x", undos)
        print_ttt_board(board)
        print_ttt_board(undos[-1])
    
    list_representation()
    tuple_representation()

def recursion_basics():
    def fib(n):
        if n == 1 or n == 2:
            return 1
        
        return fib(n - 1) + fib(n - 2)
    
    try:
        print(fib(0))
    except Exception as e:
        print(e)
    
    # print(fib(100)) # takes too long

    def fib_with_memo():
        """creates a fibonacci function that uses a memoization (not memorization!) table""" 
        memo = {}

        def fib(n):
            if n in memo:
                return memo[n]

            if n == 1 or n == 2:
                return 1
            
            ans = fib(n - 1) + fib(n - 2)
            memo[n] = ans
            return ans
        
        return fib
    
    fib = fib_with_memo()
    print(fib(100))

    def stair_climbing_with_memo():
        memo = {}

        def climb(n):
            if n in memo:
                return memo[n]
            
            if n == 0: return 0
            if n == 1: return 1
            if n == 2: return 2

            ans = climb(n - 1) + climb(n - 2)
            memo[n] = ans

            return ans
        
        return climb

    stair_climbing = stair_climbing_with_memo()
    print(stair_climbing(6))

def recursion_example_deepcopy():
    def list_deepcopy(l):
        copy = []
        for e in l:
            if isinstance(e, list):
                copy.append(list_deepcopy(e))
            else:
                copy.append(e)
        return copy
    
    L = [1, [2, [3, [4, [5]]]]]
    copy_L = list_deepcopy(L)

    L[1][1][1][1].append(6)
    print(L)
    print(copy_L)

def exception_examples():
    try:
        print(1 / 1)
    except Exception as e:
        print("error:", e)
    else:
        print("executed successfully")
    finally:
        print("always runs")

    print("---")

    try:
        print(1 / 0)
    except Exception as e:
        print("error:", e)
    else:
        print("executed successfully")
    finally:
        print("always runs")

    print("---")

    try:
        print(1 / 0)
    except ZeroDivisionError as e:
        print("ZeroDivisionError:", e)
    except Exception as e:
        print("error:", e)
    else:
        print("executed successfully")
    finally:
        print("always runs")
        
    print("---")
    
    def sqrt(x):
        if x < 0:
            raise ValueError("Square root of negative number is imaginary.")
        
        return x ** 0.5

    print(sqrt(2))
    try:
        sqrt(-1)
    except ValueError as e:
        print(e)

def assertion_examples():
    x = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    assert all(x[i] >= x[i - 1] for i in range(1, len(x))), f"{x} is not monotonically increasing"
    print("check 1 passed")

    x = [1, 2, 3, 4, 3, 6, 7, 8, 9]
    # assert all(x[i] >= x[i - 1] for i in range(1, len(x))), f"{x} is not monotonically increasing"
    # print("check 2 passed")

    def selection_sort(L):
        for i in range(len(L)):

            # find minimum value in unsorted portion of list
            min_val = None
            min_idx = None
            for j in range(i, len(L)):
                if min_idx is None or L[j] < min_val:
                    min_val = L[j]
                    min_idx = j
            
            # one-line swap
            L[i], L[min_idx] = L[min_idx], L[i]

            # after i steps, l[:i] should be sorted
            assert all(L[k] >= L[k - 1] for k in range(1, i)), f"{L[:i]} is not monotonically increasing"
    
    selection_sort(x)
    print(x)

if __name__ == "__main__":
    # list_comprehensions()
    # functions_as_arguments()
    # anonymous_functions()
    # dict_basics()
    # set_examples()
    # mutability_demo()
    # recursion_basics()
    # recursion_example_deepcopy()
    # exception_examples()
    assertion_examples()
    pass